#include "factories.h"
#include "deffmas.h"
#include "workspace.h"
#include <stdexcept>
#include <cassert>

// class Factory
WorkSpace& Factory::workSpace() {
  if (!ws) throw std::logic_error("This Factory does not have a Workspace");
  return *ws;
}

const WorkSpace& Factory::workSpace() const {
  if (!ws) throw std::logic_error("This Factory does not have a Workspace");
  return *ws;
}

Formula* Factory::copy(const Formula& formula, bool polarity) {
  switch (formula.classId()) {
  case idLiteral:
    return makeLiteral(formula.value(), polarity);
  case idJunction: {
    polarity = (formula.polarity()==polarity);
    Formula* result = makeConstant(polarity);
    for (int i=0; i<formula.count(); i++)
      result = makeJunction(result, copy(formula[i]), polarity);
    return result;
  }
  default:
    throw std::logic_error("Don't know how to copy this!");
  }
}

Formula* Factory::filter(Formula* formula, bool polarity) {
  Formula* result = 0;
  switch (formula->classId()) {
  case idLiteral:
    result = makeLiteral(formula->value(), polarity);
    break;
  case idJunction: 
    polarity = (formula->polarity()==polarity);
    result = makeConstant(polarity);
    while (formula->count())
      result = makeJunction(filter(formula->pop()), result, polarity);
    break;
  default:
    throw std::logic_error("Don't know how to filter this!");
  }
  delete formula;
  return result;
}

// class FormulaMaker
FormulaMaker::~FormulaMaker() { delete factory_; }

Factory& FormulaMaker::factory() { 
  if (!factory_) factory_ = new BasicFactory;
  return *factory_;
}

void FormulaMaker::setFactory(Factory* factory) {
  delete factory_;
  factory_ = factory;
}

// class BasicFactory
Formula* BasicFactory::makeConstant(bool polarity) { 
  return new Junction(polarity);
}

Formula* BasicFactory::makeALiteral(int value) { 
  return new Literal(value); 
}

// #include <iostream>
// #include "writers.h"
// using namespace std;

// miscellaneaous quick shorcuts
inline Formula* UseAndKill(Formula* lhs, Formula* rhs) {
  delete rhs;
//  cout << "=" << *lhs << endl;
  return lhs;
}

inline Formula* GetChildsAndKill(Formula* lhs, Formula* rhs) {
  lhs->getChilds(*rhs);
  delete rhs;
//  cout << "=" << *lhs << endl;
  return lhs;
}

inline Formula* PushAsNewChild(Formula *lhs, Formula* rhs) {
  lhs->push(rhs);
//  cout << "=" << *lhs << endl;
  return lhs;
}

Formula* BasicFactory::makeJunction(Formula* lhs, Formula* rhs, 
				    bool polarity) {
  if (!IsJunction(*lhs, polarity))
    lhs = new Junction(lhs, polarity);

  return PushAsNewChild(lhs, rhs);
}

// class SimpleFactory
Formula* SimpleFactory::makeJunction(Formula* lhs, Formula* rhs, 
	bool polarity) {
//  cout << *lhs << (polarity ? " and " : " or ") << *rhs << endl;
  if (IsConstant(*lhs, !polarity) or IsConstant(*rhs, polarity))
    return UseAndKill(lhs, rhs);
  if (IsConstant(*lhs, polarity) or IsConstant(*rhs, !polarity))
    return UseAndKill(rhs, lhs);
  if (IsJunction(*lhs, polarity)) {
    if (IsJunction(*rhs, polarity)) 
      return GetChildsAndKill(lhs, rhs);
    return PushAsNewChild(lhs, rhs);
  }
  return BasicFactory::makeJunction(rhs, lhs, polarity);
}

// class CnfFactory
Formula* CnfFactory::makeConstant(bool polarity) {
  if (polarity)
    return new Junction(true);
  else
    return new Junction(new Junction(false), true);
}

Formula* CnfFactory::makeALiteral(int value) { 
  return new Junction(new Junction(new Literal(value), false), true);
}


int FindLiteral(const Formula& clause, int lit) {
	for (int i=0; i<clause.count(); i++) {
		if (clause[i].value()==lit) return 1;
		if (clause[i].value()==-lit) return -1;
	}
	return 0;
}

Formula* CopyJoinClauses(const Formula& lhs, const Formula& rhs) {
	Formula* clause = lhs.clone();

	for (int i=0; i<rhs.count(); i++) {
		int find = FindLiteral(*clause, rhs[i].value());
		if (find<0) {
			delete clause;
			return NULL;
		}
		if (find==0) clause->push(rhs[i].clone());
	}
	return clause;
}

Formula* CnfFactory::makeJunction(Formula* lhs, Formula* rhs, bool polarity) {
  if (polarity) {
    return GetChildsAndKill(lhs, rhs);
  } else {
    Formula* result = makeTop();
    
    for (int i=0; i<lhs->count(); i++)
      for(int j=0; j<rhs->count(); j++) {
				Formula* clause = CopyJoinClauses((*lhs)[i], (*rhs)[j]);
				if (clause) result->push(clause);
      }
    
    delete lhs;
    delete rhs;
    return result;
  }
}

// class OptimizedClausalFactory
Formula* OptimizedClausalFactory::nameFormula(Formula* rhs) {
  if (rhs->count()<2) return rhs;
  int p = workSpace().signature().getFreshVariable();
  for (int i=0; i<rhs->count(); i++)
    (*rhs)[i].push(new Literal(-p));
  workSpace().addFormula(rhs, *this);
  return makeALiteral(p);
}

Formula* OptimizedClausalFactory::makeJunction(Formula* lhs, Formula* rhs,
					       bool polarity) {
  if (polarity) {
    return GetChildsAndKill(lhs, rhs);
  } else {
    if (!lhs->count()) return UseAndKill(lhs, rhs); // True or rhs -> True
    if (!rhs->count()) return UseAndKill(rhs, lhs); // lhs or True -> True
    
    lhs = nameFormula(lhs); assert(lhs->count()==1);
    rhs = nameFormula(rhs); assert(rhs->count()==1);
    
    (*lhs)[0].getChilds((*rhs)[0]);
  
    delete rhs;
    return lhs;
  }
}
