#include "options.h"
#include "utils.h"
#include <iosfwd>
#include <iomanip>
#include <sstream>
#include <algorithm>
#include <stdexcept>
#include <iostream>

using namespace std;

uint OptionsList::indent = 17;
uint OptionsList::width = 78;

// class OptionsList
void OptionsList::clear() {
  for (int i=0; i<list.size(); i++)
    delete list[i];
  list.clear();
}

void OptionsList::AddOption(const std::string& keys, const std::string& help,
  const std::string& allowed_types, const std::string& default_value,
  bool optional)
{
  if (keys=="") throw logic_error("Empty parameter key");

  Option* opt = new Option;
  opt->help = help;
  if (default_value!="") opt->help+= " (default=" + default_value + ")";
  opt->value = default_value;
  opt->optional = (allowed_types!="") and optional;
  opt->allowed_types = allowed_types;
  opt->flag = false;

  string s = keys;
  while (s!="") {
    string k = get_token(s);
    if (k=="") throw logic_error("Empty parameter key");
    if (params.find(k)!=params.end())
      throw logic_error("Duplicated parameter key: " + k);
    opt->AddKey(k);
    params[k] = opt;
  }
  list.push_back(opt);
}

void OptionsList::DumpOptionsHelp(std::ostream& os) const {
  for (int i=0; i<list.size(); i++)
    list[i]->DumpHelp(os);
}

bool OptionsList::ProcessOption(const std::string& param) {
  if (param=="") return true;
  if ((param[0]!='-') || (param=="-") || (param=="--")) return false;

  size_t p = param.find('=');
  bool has_value = (p!=string::npos);
  string name;
  string value;
  if (has_value) {
    value = param.substr(p + 1);
    name = param.substr(1, p - 1);
  } else {
    name = param.substr(1);
  }

  map<string, Option*>::iterator i = params.find(name);

  // test for error conditions
  if (i==params.end())
    throw runtime_error("Invalid command line option: " + param);

  Option& opt = *(i->second);
  if (!opt.optional and (has_value xor (opt.allowed_types!="")))
    throw runtime_error("Invalid command line option: " + param);

  if (has_value)
    opt.value = value;
  opt.flag = true;

  return true;
}

int OptionsList::ProcessOptions(int argc, char* argv[], int i) {
  while ((i < argc) && ProcessOption(argv[i])) i++;
  return i;
}

const std::string& OptionsList::operator[](const std::string& index) const {
  map<string, Option*>::const_iterator i = params.find(index);
  if (i==params.end()) throw logic_error("Unknown option: " + index);
  return i->second->value;
}

bool OptionsList::flag(const std::string& index) const {
  map<string, Option*>::const_iterator i = params.find(index);
  if (i==params.end()) throw logic_error("Unknown option: " + index);
  return i->second->flag;
}

// class OptionsList::Option
void OptionsList::Option::AddKey(const string& key) {
  string s;
  if (allowed_types!="") {
    s = "=" + allowed_types;
    if (optional) s = "[" + s + "]";
  }
  nice_append(prototype, "-" + key + s);
}

void OptionsList::Option::DumpHelp(std::ostream& os) const {
  os << "  " << setw(indent - 3) << left << prototype;
  uint c = max(prototype.length() + 2, indent - 1);
  if (c >= indent) {
    os << endl;
    for (int i=1; i<indent; i++) os << " ";
    c = indent - 1;
  }
  istringstream is(help);
  string word;
  while (is >> word) {
    c+= word.length() + 1;
    if (c > width) {
      os << endl;
      for (int i=1; i<indent; i++) os << " ";
      c = indent + word.length();
    }
    os << " " << word;
  }
  os << endl;
}
