#ifndef REGPROP_H
#define REGPROP_H

#include "workspace.h"
#include "factories.h"
#include "generators.h"
#include "writers.h"
#include "formulas.h"
#include "options.h"
#include "utils.h"
#include <stdexcept>
#include <iostream>
#include <cstdlib>

using namespace std;

ostream& operator<<(ostream& os, const RegShape& shape) {
  for (int i=0; i<shape.size(); i++) {
    if (i) os << "-";
    os << shape[i];
  }
}

class RegularGenerator {
  string description;
  uint variables;
  uint count;
  uint seed;
  int translation;
  RegShape shape;
 protected:
  void RegisterOptions(OptionsList& p);
  void ProcessOptions(int argc, char* args[]);
  Factory* getFactory(WorkSpace& workSpace);
 public:
  void Execute(int argc, char* args[]);
};

void RegularGenerator::Execute(int argc, char* args[]) {
  ProcessOptions(argc, args);

  WorkSpace ws(description, new FixedSignature(variables));

  srand(seed);
  RegularRandom generator(shape, ws, getFactory(ws));
  ws.addFormula(generator.makeFormulas(count), generator.factory());

/*
  int space = 0, maxlen = 0;
  for (int i=0; i<ws.program().count(); i++) {
    int len = ws.program()[i].count();
    space+= len;
    maxlen = max(maxlen, len);
  }
  cout << "c fresh=" << ws.signature().systemVariables() 
       << " space=" << space << " maxlen=" << maxlen << endl;  */
       
  cout << ws;
}

Factory* RegularGenerator::getFactory(WorkSpace& workSpace) {
  switch (translation) {
  case 0:
    return new OptimizedClausalFactory(workSpace);
  case 1:
    return new CnfFactory(workSpace);
  default:
    return new SimpleFactory(workSpace);
  }
}

void RegularGenerator::ProcessOptions(int argc, char* args[]) {
  OptionsList p;

  RegisterOptions(p);
  int i = p.ProcessOptions(argc, args);

  if ((i>=argc) || p.flag("help")) {
    cout << "Usage: " << args[0] << " [options] <shape>" << endl;
    cout << "Generates random non-clausal formulas with balanced shapes." << endl;
    cout << endl << "Available optoins:" << endl << p;
    exit(0);
  }

  string the_shape = args[i++];
  while (i<argc) (the_shape+= " ")+= args[i++];

  shape.clear();
  while (the_shape!="") {
    string token = get_token(the_shape, "-, /");
    if (token!="")
      shape.push_back(unstring<uint>(token));
  }

  variables = unstring<uint>(p["variables"]);
  description = "-v=" + stringify(variables);
  
  if (p.flag("ratio")) {
    double ratio = unstring<double>(p["ratio"]);
    description+= " -r=" + stringify(ratio);
    count = (uint)(ratio*variables);
  } else {
    count = unstring<uint>(p["count"]);
    description+= " -c=" + stringify(count);
  }
  
  seed = (p.flag("seed") ? unstring<uint>(p["seed"]) : randseed());
  description+= " -s=" + stringify(seed);

  translation = select(p["trans"], "OPT|STD|NONE");
  if (translation < 0) {
    cerr << "Transformation `" << p["trans"] << "' unknown. " << endl
	 << "Try OPT, STD or NONE." << endl;
    exit(1);
  }

  int format = select(p["format"], "CNF|NONCNF|ISCAS|SEXPR|SAT|PROP");
  if ((translation==2) and (format==0)) {
    translation = 0;
    description+= " -t=OPT";
  } else {
    description+= " -t=" + p["trans"];
  }
  description+= " -f=" + p["format"];
  description+= " \"" + stringify(shape) + "\"";
  
  switch (format) {
  case 0:
    systemWriter.setWriter(new CnfWriter);
    break;
  case 1:
    systemWriter.setWriter(new NoncnfWriter);
    break;
  case 2:
    systemWriter.setWriter(new IscasWriter);
    break;
  case 3:
    systemWriter.setWriter(new SexprWriter);
    break;
  case 4:
    systemWriter.setWriter(new SatWriter);
    break;
  case 5:
    systemWriter.setWriter(new BasicWriter);
    break;
  default:
    cerr << "Format `" << p["format"] << "' unknown. " << endl
         << "Try CNF, NONCNF, ISCAS, SEXPR, SAT, or PROP." << endl;
    exit (1);
  }
}

void RegularGenerator::RegisterOptions(OptionsList& p) {
  p.AddOption("h|help", "Show this help text and exit.");
  p.AddOption("v|variables", "Max number of variables to use.",
	      "<num>", "100");
  p.AddOption("c|count", "Number of formulas to generate.",
	      "<num>", "1");
  p.AddOption("r|ratio", "Ratio of formulas/variables. It is equivalent "
	      "to set a count of [ratio*variables] formulas. If a ratio "
	      "is specified the value of count is ignored.",
	      "<real>");
  p.AddOption("f|format", 
	      "Format used to output the formula: "
              "CNF - Standard DIMACS CNF format (you must also specify a "
              "clausal form translation), "
              "NONCNF - non-CNF DIMACS style (F. Bacchus and T. Walsh), "
              "ISCAS - circuit verification ISCAS format, "
              "SEXPR - LISP style formula, "
              "SAT - Original suggested DIMACS SAT format, "
              "PROP - Formula in infix notation.",
	      "<str>", "PROP");
  p.AddOption("t|trans", "Translation applied to the formula: "
	      "OPT - Optimized Structure Preserving Clausal Form, "
	      "STD - Standard (exponential) CNF translation, "
	      "NONE - Do not perform any clausal form translation.",
	      "<str>", "NONE");
  p.AddOption("s|seed", "Seed used for the random number generator. "
	      "The default uses the system clock and process id.",
	      "<num>");
}

#endif // REGPROP_H
