#include "writers.h"
#include "formulas.h"
#include "workspace.h"
#include <sstream>
#include <cassert>
#include <vector>

using namespace std;

SystemWriter systemWriter;

std::ostream& operator<<(std::ostream& os, const Formula& formula) {
  systemWriter.writer().formulaToStream(formula, os);
  
  return os;
}

std::ostream& operator<<(std::ostream& os, const WorkSpace& workspace) {
  systemWriter.writer().workspaceToStream(workspace, os);

  return os;
}

// class SystemWriter
Writer& SystemWriter::writer() {
  if (!writer_) writer_ = new BasicWriter;
  return *writer_;
}

void SystemWriter::setWriter(Writer* writer) {
  delete writer_;
  writer_ = writer;
}

// class BasicWriter
void BasicWriter::formulaToStream(const Formula& formula, ostream& os) {
  switch (FormulaId id = formula.classId()) {
  case idLiteral:
    os << formula.value();
    break;
  case idJunction:
    if (formula.count()==0) {
      os << (formula.polarity() ? "top" : "bot");
    } else if (formula.count()==1) {
      formulaToStream(formula[0], os);
    } else {
      os << "(";
      formulaToStream(formula[0], os);
      for (int i=1; i<formula.count(); i++) {
	os << (formula.polarity() ? " and " : " or ");
	formulaToStream(formula[i], os);
      }
      os << ")";
    }
    break;
  default:
    os << "[id:" << id << "?]";
  }
}

void BasicWriter::workspaceToStream(const WorkSpace& workspace, ostream& os) {
  if (workspace.description()!="")
    os << "c " << workspace.description() << endl;
  formulaToStream(workspace.program(), os);
  os << endl;
}

// class CnfWriter
void CnfWriter::formulaToStream(const Formula& formula, ostream& os) {
  assert(IsConjunction(formula));
  for (int i=0; i<formula.count(); i++) {
    assert(IsDisjunction(formula[i]));
    for (int j=0; j<formula[i].count(); j++) {
      assert(IsLiteral(formula[i][j]));
      os << formula[i][j].value() << " ";
    }
    os << "0" << endl;
  }
}

void CnfWriter::workspaceToStream(const WorkSpace& workspace, ostream& os) {
  if (workspace.description()!="")
    os << "c " << workspace.description() << endl;

  os << "p cnf " << workspace.signature().variables() << " "
     << workspace.program().count() << endl;
  formulaToStream(workspace.program(), os);
}

// class SatWriter
void SatWriter::formulaToStream(const Formula& formula, ostream& os) {
  switch (FormulaId id = formula.classId()) {
  case idLiteral:
    os << formula.value();
    break;
  case idJunction:
    os << (formula.polarity() ? "*" : "+");
    os << "(";
    for (int i=0; i<formula.count(); i++) {
      if (i) os << " ";
      formulaToStream(formula[i], os);
    }
    os << ")";
    break;
  default:
    os << "[id:" << id << "?]";
  }
}

void SatWriter::workspaceToStream(const WorkSpace& workspace, ostream& os) {
  if (workspace.description()!="")
    os << "c " << workspace.description() << endl;

  os << "p sat " << workspace.signature().variables() << endl;
  os << "(";
  formulaToStream(workspace.program(), os);
  os << ")" << endl;
}

// class SexprWriter
void SexprWriter::formulaToStream(const Formula& formula, ostream& os) {
  switch (FormulaId id = formula.classId()) {
  case idLiteral:
    if (formula.value() < 0) {
      os << "(NOT " << -formula.value() << ")";
    } else {
      os << formula.value();
    }
    break;
  case idJunction:
    os << "(";
    os << (formula.polarity() ? "AND " : "OR ");
    for (int i=0; i<formula.count(); i++) {
      if (i) os << " ";
      formulaToStream(formula[i], os);
    }
    os << ")";
    break;
  default:
    os << "[id:" << id << "?]";
  }
}

void SexprWriter::workspaceToStream(const WorkSpace& workspace, ostream& os) {
//  if (workspace.description()!="")
//    os << "c " << workspace.description() << endl;
//
//  os << "p sat " << workspace.signature().variables() << endl;
  formulaToStream(workspace.program(), os);
  os << endl;
}

// class NoncnfWriter
void NoncnfWriter::count(const Formula& formula, int& node) {
  if (formula.classId()==idLiteral) return;
  for (int i=0; i<formula.count(); i++)
    count(formula[i], node);
  node++;
}

int NoncnfWriter::visit(const Formula& formula, ostream& os, int& node) {
  switch (FormulaId id = formula.classId()) {
  case idLiteral:
    return formula.value();
  default:
    std::vector<int> childs;
    for (int i=0; i<formula.count(); i++) {
      childs.push_back(visit(formula[i], os, node));
    }
    node++;
    os << (formula.polarity() ? "4 -1 " : "6 -1 ") << node;
    for (int i=0; i<formula.count(); i++) {
      os << " " << childs[i];
    }
    os << " 0" << endl;
    return node;
  }
}

void NoncnfWriter::formulaToStream(const Formula& formula, ostream& os) {
  assert( false );
}

void NoncnfWriter::workspaceToStream(const WorkSpace& workspace, ostream& os) {
  int vars = workspace.signature().variables();
  int node = vars;
  count(workspace.program(), vars);
  if (workspace.description()!="")
    os << "c " << workspace.description() << endl;
  os << "p noncnf " << vars << endl;
  visit(workspace.program(), os, node);
}


// class IscasWriter
int IscasWriter::visit(const Formula& formula, ostream& os, int &node) {
  bool output;

  if (formula.classId()==idLiteral) return 0;
  
  output = (node<0);
  if (output) { node = 0; }
  std::vector<int> childs;
  for (int i=0; i<formula.count(); i++) {
    childs.push_back(visit(formula[i], os, node));
  }
  node++;
  if (output) os << "out"; else os << "g_" << node;
  os << " = " << (formula.polarity() ? "AND(" : "OR(");
  for (int i=0; i<formula.count(); i++) {
    if (i) os << ", ";
    if (childs[i]) {
      os << "g_" << childs[i];
    } else {
      int v = formula[i].value();
      if (v > 0) os << "v_" << v; else os << "n_" << -v;
    }
  }
  os << ")" << endl;
  return node;
}

void IscasWriter::formulaToStream(const Formula& formula, ostream& os) {
  int node = -1;
  visit(formula, os, node);
}

void IscasWriter::workspaceToStream(const WorkSpace& workspace, ostream& os) {
  if (workspace.description()!="")
    os << "# " << workspace.description() << endl;

  for (int i=1; i<=workspace.signature().variables(); i++) {
    os << "INPUT(v_" << i << ")" << endl;
  }
  os << "OUTPUT(out)" << endl;
  for (int i=1; i<=workspace.signature().variables(); i++) {
    os << "n_" << i << " = NOT(v_" << i << ")" << endl;
  }

  int node = -1;
  visit(workspace.program(), os, node);
}

